/**
 * @file    mod_auta.c
 * @author  Paco Rodríguez (prodrig@disca.upv.es), adaptado de Derek Molloy
 * @date    20 Octubre 2017
 * @version 0.2
 * @brief   Character device driver como módulo externo del kernel de Linux de carga dinámina (LKM)
 * 
 * Segunda versión de módulo LKM, esta vez sí se aceptan peticiones de programas de usuario:
 * - "abrir" driver
 * - "cerrar" driver
 * - escribir datos en el driver
 * - leer datos del driver
 * 
 * El módulo exporta dos parámetros de funcionamiento que se pueden modificar
 * bien durante la carga del módulo o a posteriori, una vez cargado.
 * 
 * Parámetros:
 * - nombre: es el nombre del módulo en los ficheros de log (valor por defecto "mod_auta")
 * - valor: un valor numérico que aparece en los mensajes de log (valor por defecto 99)
 * 
 * Los mensajes de depuración van a /var/log/kern.log (hacer tail del fichero)
 * 
 * @see http://www.derekmolloy.ie/ para obtener el fichero original
*/
#include <linux/init.h>             // Macros used to mark up functions e.g., __init __exit
#include <linux/module.h>           // Core header for loading LKMs into the kernel
#include <linux/kernel.h>           // Contains types, macros, functions for the kernel
#include <asm/uaccess.h>            // copy_from_user/copy_to_user
#include <linux/device.h>           // Header to support the kernel Driver Model
#include <linux/fs.h>               // Header for the Linux file system support

#define  DEVICE_NAME "mod_auta"     ///< The device will appear at /dev/mod_auta using this value
#define  CLASS_NAME  "auta"         ///< The device class -- this is a character device driver


#define MESSAGE_MAX_LEN     200
 
 
/**
 * @brief Información que aparecerá reflejada con modinfo, además de los parámetros
 */
MODULE_LICENSE("GPL");              ///< The license type -- this affects runtime behavior
MODULE_AUTHOR("Tu nombre aquí");    ///< The author -- visible when you use modinfo
MODULE_DESCRIPTION("A simple Linux driver.");  ///< The description -- see modinfo
MODULE_VERSION("0.2");              ///< The version of the module

/**
 * @brief Parámetro 'nombre' del módulo
 * 
 * La macro module_param() es la que convierte la variable nombre en un parámetro del módulo.
 * Indica el nombre (nombre, sin comillas), el tipo (charp --> char *), y los permisos
 * (S_IRUGO --> Se puede leer pero no modificar)
 * 
 * Permisos de acceso (en linux/stat.h):
 * #define S_IRWXUGO	(S_IRWXU|S_IRWXG|S_IRWXO)
 * #define S_IALLUGO	(S_ISUID|S_ISGID|S_ISVTX|S_IRWXUGO)
 * #define S_IRUGO		(S_IRUSR|S_IRGRP|S_IROTH)
 * #define S_IWUGO		(S_IWUSR|S_IWGRP|S_IWOTH)
 * #define S_IXUGO		(S_IXUSR|S_IXGRP|S_IXOTH)
 * 
 * donde los valores individuales son (de sys/stat.h):
 * S_IRUSR - Read permission bit for the owner of the file. 
 * S_IWUSR - Write permission bit for the owner of the file. 
 * S_IXUSR - Execute (for ordinary files) or search (for directories) permission bit for the owner of the file. 
 * S_IRWXU - This is equivalent to ‘(S_IRUSR | S_IWUSR | S_IXUSR)’.
 * S_IRGRP - Read permission bit for the group owner of the file.
 * S_IWGRP - Write permission bit for the group owner of the file. 
 * S_IXGRP - Execute or search permission bit for the group owner of the file.
 * S_IRWXG - This is equivalent to ‘(S_IRGRP | S_IWGRP | S_IXGRP)’.
 * S_IROTH - Read permission bit for other users.
 * S_IWOTH - Write permission bit for other users.
 * S_IXOTH - Execute or search permission bit for other users.
 * S_IRWXO - This is equivalent to ‘(S_IROTH | S_IWOTH | S_IXOTH)’.
 * S_ISUID - This is the set-user-ID on execute bit.
 * S_ISGID - This is the set-group-ID on execute bit.
 * S_ISVTX - This is the sticky bit.
 * 
 * ¡OJO!: Permitir que cualquiera escriba el parámetro *no* está permitido
 * (mensaje de error ... BUILD_BUG_ON_ZERO((perms) & 2) ...) 
 * 
 * La macro MODULE_PARM_DESC() permite añadir un texto descriptivo para el parámetro
 */
static char *nombre = "mod_auta";     ///< Ejemplo de parámetro para el LKM -- default value "mod_auta"
module_param(nombre, charp, S_IRUGO); ///< Param desc.: charp = char ptr, S_IRUGO --> any can read/nobody can write
MODULE_PARM_DESC(nombre, " The name to display in /var/log/kern.log");  ///< parameter description

/**
 * @brief Parámetro 'valor' del módulo
 * 
 * La macro module_param() es la que convierte la variable valor en un parámetro del módulo.
 * Indica el nombre (valor, sin comillas), el tipo (int), y los permisos
 * (S_IRWXU --> El propietario puede leer y escribir +
 *  S_IRGRP --> El grupo del propietario puede leer +
 *  S_IROTH --> Cualquiera puede leer
 * )
 * Para ver otros permisos de acceso, mira la descripción de @ref nombre
 * 
 * ¡OJO!: Permitir que cualquiera escriba el parámetro *no* está permitido
 * (mensaje de error ... BUILD_BUG_ON_ZERO((perms) & 2) ...) 
 * 
 * La macro MODULE_PARM_DESC() permite añadir un texto descriptivo para el parámetro
 */
static int valor = 99;     ///< Ejemplo de parámetro para el LKM -- default value 99
module_param(valor, int, S_IRWXU | S_IRGRP | S_IROTH); ///< Param desc.: int, S_IRWXUGO --> any can read/write
MODULE_PARM_DESC(valor, " The value to display in /var/log/kern.log");  ///< parameter description

/**
 * @brief Datos internos del módulo
 */
struct dev_info_s {
    /* Información común para cualquier driver de este tipo */
    int    majorNumber;                     ///< Stores the device number -- determined automatically
    struct class*  clase;                   ///< The device-driver class struct pointer
    struct device* disp;                    ///< The device-driver device struct pointer
    /* Información específica para este driver */
    char message[MESSAGE_MAX_LEN];      ///< Buffer de memoria con una copia del mensaje del programa de usuario
    int message_len;                    ///< Longitud del mensaje del programa de usuario
    int number_opens_total;             ///< Número total de veces que se ha abierto el driver
    int number_opens_current;           ///< Número actual de veces que se ha abierto el driver
};
typedef struct dev_info_s dev_info_t;
static dev_info_t dev_info = {
    .majorNumber = 0, .clase = NULL, .disp = NULL,
    .message = {0}, .message_len = 0, .number_opens_total = 0, .number_opens_current = 0
};


/*
 * Prototipo de las functiones que responderán a las peticiones de los programas de usuario
 */
static int     dev_open(struct inode *, struct file *);
static int     dev_release(struct inode *, struct file *);
static ssize_t dev_read(struct file *, char *, size_t, loff_t *);
static ssize_t dev_write(struct file *, const char *, size_t, loff_t *);


/** 
 * @brief Tabla que relaciona operaciones de los programas de usuario y funciones del driver. 
 * 
 * The file_operations structure from /linux/fs.h lists the callback functions that 
 * you wish to associated with your file operations using a C99 syntax structure. 
 * char devices usually implement open, read, write and release calls
 * 
 * @note Sólo usamos de esta estructura los campos que nos interesan.
 * Eso significa que el resto de operaciones posibles no están disponibles
 */
static struct file_operations fops = {
   .open = dev_open,
   .read = dev_read,
   .write = dev_write,
   .release = dev_release
};

/* -----------------------------------------------------------------------------
 * Operaciones soportadas por el driver
 * -----------------------------------------------------------------------------
 */
 
/** 
 * @brief The device open function that is called each time the device is opened
 * 
 * Incrementamos los contadores de veces que se ha abierto 
 * 
 * @param inodep A pointer to an inode object (defined in linux/fs.h)
 * @param filep A pointer to a file object (defined in linux/fs.h)
 * 
 */
static int dev_open(struct inode *inodep, struct file *filep)
{
    dev_info.number_opens_current++;
    dev_info.number_opens_total++;
    pr_debug("[%s] dev_open: Actualmente abierto por %d procesos (%d en total durante la vida del driver)\n", 
                nombre, dev_info.number_opens_current, dev_info.number_opens_total);
    /* Éxito */
    return 0;
}

/** 
 * @brief The device release function that is called whenever the device is closed/released by the userspace program
 * 
 * Decrementamos el contador de veces que está actualmente abierto
 * 
 * @param inodep A pointer to an inode object (defined in linux/fs.h)
 * @param filep A pointer to a file object (defined in linux/fs.h)
 */
static int dev_release(struct inode *inodep, struct file *filep)
{
    if (dev_info.number_opens_current > 0) {
        dev_info.number_opens_current--;
    }
    pr_debug("[%s] dev_close: Actualmente abierto por %d procesos (%d en total durante la vida del driver)\n", 
                nombre, dev_info.number_opens_current, dev_info.number_opens_total);
    /* Éxito */
    return 0;
}
  

/** 
 * @brief This function is called whenever device is being read from user space 
 * 
 * Un programa de usuario nos reclama un mensaje (se supone que previamente nos ha enviado algo)
 * 
 * @param filep A pointer to a file object (defined in linux/fs.h)
 * @param buffer The pointer to the buffer to which this function writes the data
 * @param len The length of the b
 * @param offset The offset if required
 * 
 * @warning Es muy importante usar copy_to_user porque el buffer pertenece al 
 * espacio de direcciones del programa de usuario, no podemos acceder directamente
 */
static ssize_t dev_read(struct file *filep, char *buffer, size_t len, loff_t *offset)
{
    int result = 0;

/*    
    if (dev_info.message_len == 0) {
        pr_debug("[%s] dev_read: Actualmente no hay mensaje que enviar!\n", nombre);
        return -1;
    }
*/
    if (dev_info.message_len > len) {
        pr_debug("[%s] dev_read: El buffer del usuario es demasiado pequeño (%d bytes, el mensaje necesita %d bytes)!\n", 
                    nombre, len, dev_info.message_len);
        return -2;
    }
   
    // copy_to_user has the format (*to - user space, *from - kernel space, size) and returns 0 on success
    // (result is the number of *non* copied bytes)
    result = copy_to_user(buffer, dev_info.message, dev_info.message_len);

    if (result == 0) {               // if true then have success
        pr_debug("[%s] dev_read: Enviados %d datos al buffer del usuario!\n", 
                    nombre, dev_info.message_len);
        /* Hacemos como que ya no tenemos mensaje almacenado */
        dev_info.message_len = 0;
    } else {
        pr_debug("[%s] dev_read: Error al enviar %d datos al buffer del usuario (%d datos restantes)!\n", 
                    nombre, dev_info.message_len, result);
        result = -EFAULT;              // Failed -- return a bad address message (i.e. -14)
    }
    
    /* 0 = Éxito */
    return result;
}
 
/** 
 * @brief This function is called whenever the device is being written to from user space
 * 
 * Un programa de usuario nos envía un mensaje (se supone que en el futuro nos lo reclamará)
 * 
 * @param filep A pointer to a file object
 * @param buffer The buffer to that contains the string to write to the device
 * @param len The length of the array of data that is being passed in the const char buffer
 * @param offset The offset if required
 * 
 * @warning Es muy importante usar copy_from_user porque el buffer pertenece al 
 * espacio de direcciones del programa de usuario, no podemos acceder directamente
 */
static ssize_t dev_write(struct file *filep, const char *buffer, size_t len, loff_t *offset)
{
    int result = 0;

/*    
    if (dev_info.message_len != 0) {
        pr_debug("[%s] dev_write: Actualmente YA hay un mensaje almacenado!\n", nombre);
        return -1;
    }
*/
    if (MESSAGE_MAX_LEN < len) {
        pr_debug("[%s] dev_write: El buffer del usuario es demasiado grande (%d bytes, el mensaje máximo es de %d bytes)!\n", 
                    nombre, len, MESSAGE_MAX_LEN);
        return -2;
    }
   
    // copy_to_user has the format (*to - kernel space, *from - user space, size) and returns 0 on success
    // (result is the number of *non* copied bytes)
    result = copy_from_user(dev_info.message, buffer, len);
 
    if (result == 0) {               // if true then have success
        pr_debug("[%s] dev_write: %d datos recibiudos desde el buffer del usuario!\n", 
                    nombre, len);
        dev_info.message_len = len;
    } else {
        pr_debug("[%s] dev_write: Error al recibir %d datos desde el buffer del usuario (%d datos restantes)!\n", 
                    nombre, len, result);
        /* Hacemos como que no hemos recibido ningún mensaje */
        dev_info.message_len = 0;
        result = -EFAULT;              // Failed -- return a bad address message (i.e. -14)
    }
    
    /* 0 = Éxito */
    return len;
}
 

/* -----------------------------------------------------------------------------
 * Inicialización y descarga del driver
 * -----------------------------------------------------------------------------
 */
 
 
/** 
 * @brief The LKM initialization function
 * 
 * The static keyword restricts the visibility of the function to within this C file.
 * The __init macro means that for a built-in driver (not a LKM) 
 * the function is only used at initialization time 
 * and that it can be discarded and its memory freed up after that point.
 * 
 * En este caso la inicialización es mucho más compleja ya que hay que
 * obtener un identificador de dispositivo que no esté en uso, entre otras cosas,
 * para poder registrar el dispositivo en /dev
 * 
 * @return returns 0 if successful
 */
static int __init mod_auta_init(void)
{
    pr_debug("[%s] init: Hello from the LKM! Current value is %d\n", nombre, valor);
    
    // Paso 1 init: Obtener un major number y registrar las operaciones soportadas
    dev_info.majorNumber = register_chrdev(0, DEVICE_NAME, &fops);
    if (dev_info.majorNumber<0) {
        pr_debug("[%s] init: register_chrdev error!\n", nombre);
        return dev_info.majorNumber;
    }
    
    // Paso 2 init: Registrar la clase de dispositivo (retorna un puntero, usar IS_ERR())
    dev_info.clase = class_create(THIS_MODULE, CLASS_NAME);
    if (IS_ERR(dev_info.clase)) {                   // Check for error and clean up if there is
        pr_debug("[%s] init: class_create error!\n", nombre);
        // Deshacer paso 1 antes de terminar
        unregister_chrdev(dev_info.majorNumber, DEVICE_NAME);
        // El código de error está almacenado en el puntero clase,
        // usar PTR_ERR() para obtener el valor numérico
        return PTR_ERR(dev_info.clase);             // Correct way to return an error on a pointer
    }
    
    // Paso 3 init: Registrar el driver de dispositivo (retorna un puntero, usar IS_ERR())
    dev_info.disp = device_create(dev_info.clase, NULL, MKDEV(dev_info.majorNumber, 0), NULL, DEVICE_NAME);
    if (IS_ERR(dev_info.disp)) {                    // Clean up if there is an error
        pr_debug("[%s] init: device_create error!\n", nombre);
        // Deshacer paso 2 antes de terminar, después
        class_destroy(dev_info.clase);
        // Deshacer paso 1 antes de terminar
        unregister_chrdev(dev_info.majorNumber, DEVICE_NAME);
        // El código de error está almacenado en el puntero disp,
        // usar PTR_ERR() para obtener el valor numérico
        return PTR_ERR(dev_info.disp);
    }

    /* Éxito */
    return 0;
}

/** 
 * @brief The LKM cleanup function
 * 
 * Similar to the initialization function, it is static. 
 * The __exit macro notifies that if this code is used for a built-in driver (not a LKM) 
 * that this function is not required.
 */
static void __exit mod_auta_exit(void)
{
    // Deshacer primero el paso 3 de init,
    device_destroy(dev_info.clase, MKDEV(dev_info.majorNumber, 0));     
    // Después deshacer el paso 2 de init,
    class_destroy(dev_info.clase);
    // Y finalmente deshacer el paso 1 de init,
    unregister_chrdev(dev_info.majorNumber, DEVICE_NAME);
    pr_debug("[%s] exit: Goodbay from the LKM! Current value is %d\n", nombre, valor);
}

/** @brief A module must use the module_init() module_exit() macros from linux/init.h
 * 
 * El uso de las macros module_init() y module_exit() es obligatorio para que el sistema
 * sepa qué funciones corresponden a la inicialización (carga del LKM) y finalización
 * (descarga del LKM) del módulo.
 */
module_init(mod_auta_init);
module_exit(mod_auta_exit);
