/**
 * @file   test_mod_auta.c
 * @author Paco Rodríguez (prodrig@disca.upv.es), adaptado de Derek Molloy
 * @date   20 Octubre 2017
 * @version 0.1
 * @brief  Programa de test del driver mod_auta versión 0.2 (character device driver)
 * 
 * Este es un sencillo programa del espacio de usuario que accede a los servicios del driver.
 * Recuerda que el driver se monta en /dev/mod_auta.
 * 
 * @see http://www.derekmolloy.ie/  para obtener el fichero original
*/
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/time.h>

const char *device = "/dev/mod_auta";


#define MSG_WRITE_MAXLEN    100
#define MSG_READ_MAXLEN     100

char msg_write[MSG_WRITE_MAXLEN];
int msg_write_len;

char msg_read[MSG_READ_MAXLEN];
int msg_read_len;

int getche(void);

int init_random_generator(void)
{
    char errmsg[100];
    int result = 1;
    
    /* 
     * Inicializar semilla del generador de números aleatorios
     * con los microsegundos de la hora actual
     */
    struct timeval t;
    if (gettimeofday(&t, 0) < 0) {
        /* Imprimir mensaje de error si algo va mal */
        snprintf(errmsg, 100, "[init_random_generator] Error al hacer gettimeofday()");
        perror (errmsg);
        result = 0;
    }
    srandom(t.tv_usec);
    return result;
}

int dev_open(const char *nombre) 
{
    int fd;
    char errmsg[100];
    
    /* Abrir el driver para lectura/esccritura */
    fd = open (nombre, O_RDWR);
    if (fd < 0) {
        /* Imprimir mensaje de error si algo va mal */
        snprintf(errmsg, 100, "[dev_open] Error al abrir %s", nombre);
        perror (errmsg);
    }
    return fd;
}

void dev_close(const char *nombre, int fd)
{
    char errmsg[100];
    if (close(fd) < 0) {
        /* Imprimir mensaje de error si algo va mal */
        snprintf(errmsg, 100, "[dev_close] Error al cerrar %s (handle %d)", nombre, fd);
        perror (errmsg);
    }
}

    
    


int main(int argc, char *argv[])
{
    int result = 0, valor, fd;
    int operations_total = 0, operations_failed = 0;
    char c;
    
    if (!init_random_generator()) {
        return -1;
    }
    
    fd = dev_open(device);
    if (fd < 0){
        return errno;
    }
    
    valor = rand();
    snprintf(msg_write, MSG_WRITE_MAXLEN, "Dato aleatorio (%d) que se envía al driver", valor);
    msg_write_len = strlen(msg_write);

	system("/bin/stty raw");

    while (1) {
        printf("Press q/Q to finish\r\n");
        printf("Press any other key to 1) write then 2) read from the device\r\n");
        
        c = tolower(getche());
        if (c == 'q') {
            result = 0;
            break;
        }
        
        operations_total++;
        
        /*
         * Escribir siempre el mismo mensaje (diferente en cada ejecución)
         */
        result = write(fd, msg_write, msg_write_len);
        if (result < 0){
            system("/bin/stty cooked");
            perror("\tStep 1: failed to write");
            operations_failed++;
            break;
        }
        
        /*
         * Intercalamos un tiempo indefinido entre escritura y lectura subsiguiente
         */
        printf("\tStep 1 finished (device written), this is the message sent:\r\n\t<%s>\r\n", msg_write);
        printf("\tpress any key to 2) read from the device\r\n");
        getche();
        
        /*
         * Leer el mensaje devuelto por el driver
         */
        result = read(fd, msg_read, MSG_READ_MAXLEN);
        if (result < 0) {
            system("/bin/stty cooked");
            perror("\tStep 2: failed to read");
            operations_failed++;
            break;
        }
        printf("\tStep2 finished (device read), this is the received message:\r\n\t<%s>\r\n", msg_read);
        if (strncmp(msg_write, msg_read, msg_write_len) != 0) {
            operations_failed++;
            printf("\tStep 2: comparison failed!\r\n");
        } else {
            printf("\tStep 2 finished ok\r\n");
        }
        
    }

	system("/bin/stty cooked");
    dev_close(device, fd);
    printf ("Closing (error code %d), total operations %d (%d failed)\n", result, operations_total, operations_failed);
    return result;
}


/* getchar() sin eco, código original de Edward A Kimble (https://stackoverflow.com/questions/558009/ansi-c-no-echo-keyboard-input) */
int getche(void)
{
    struct termios oldattr, newattr;
    int ch;
    tcgetattr( STDIN_FILENO, &oldattr );
    newattr = oldattr;
    newattr.c_lflag &= ~( ICANON | ECHO);//knock down keybuffer
    tcsetattr( STDIN_FILENO, TCSANOW, &newattr );
    system("stty -echo");//shell out to kill echo
    ch = getchar();
    system("stty echo");
    tcsetattr( STDIN_FILENO, TCSANOW, &oldattr );
    return ch;
}
